@@ -6,29 +6,29 @@ This module is a collection of useful building-blocks for hacking the Python
 "with" statement.  It combines ideas from several neat with-statement hacks 
 I found around the internet into a suite of re-usable components:
 
-    http://www.mechanicalcat.net/richard/log/Python/Something_I_m_working_on.3
-    http://billmill.org/multi_line_lambdas.html
-    http://code.google.com/p/ouspg/wiki/AnonymousBlocksInPython
+  * http://www.mechanicalcat.net/richard/log/Python/Something_I_m_working_on.3
+  * http://billmill.org/multi_line_lambdas.html
+  * http://code.google.com/p/ouspg/wiki/AnonymousBlocksInPython
 
-By subclasses the appropriate context managers from this module, you can
+By subclassing the appropriate context managers from this module, you can
 easily do things such as:
 
-    * skip execution of the code inside the with-statement
-    * set local variables in the frame executing the with-statement
-    * capture the bytecode from inside the with-statement
-    * capture local variables defined inside the with-statement
+  * skip execution of the code inside the with-statement
+  * set local variables in the frame executing the with-statement
+  * capture the bytecode from inside the with-statement
+  * capture local variables defined inside the with-statement
 
 Building on these basic tools, this module also provides some useful prebuilt
 hacks:
 
-    xargs:      call a function with additional arguments defined in the
-                body of the with-statement
-    xkwargs:    call a function with additional keyword arguments defined
-                in the body of the with-statement
-    namespace:  direct all variable accesses and assignments to the attributes
-                of a given object (like "with" in JavaScript or VB)
-    keyspace:   direct all variable accesses and assignments to the keys of
-                of a given object (like namespace() but for dicts)
+  :xargs:      call a function with additional arguments defined in the
+               body of the with-statement
+  :xkwargs:    call a function with additional keyword arguments defined
+               in the body of the with-statement
+  :namespace:  direct all variable accesses and assignments to the attributes
+               of a given object (like "with" in JavaScript or VB)
+  :keyspace:   direct all variable accesses and assignments to the keys of
+               of a given object (like namespace() but for dicts)
 
 WithHacks makes extensive use of Noam Raphael's fantastic "byteplay" module;
 since the official byteplay distribution doesn't support Python 2.6, a local
@@ -40,7 +40,7 @@ from __future__ import with_statement
 
 __ver_major__ = 0
 __ver_minor__ = 1
-__ver_patch__ = 0
+__ver_patch__ = 1
 __ver_sub__ = ""
 __version__ = "%d.%d.%d%s" % (__ver_major__,__ver_minor__,
                               __ver_patch__,__ver_sub__)
@@ -48,7 +48,6 @@ __version__ = "%d.%d.%d%s" % (__ver_major__,__ver_minor__,
 import sys
 import new
 import copy
-from dis import HAVE_ARGUMENT
 try:
     import threading
 except ImportError:
@@ -56,6 +55,7 @@ except ImportError:
 
 
 from withhacks.byteplay import *
+from withhacks.frameutils import inject_trace_func, extract_code
 
 
 class _ExitContext(Exception):
@@ -63,53 +63,16 @@ class _ExitContext(Exception):
     pass
 
 
+def _exit_context(frame):
+    """Simple function to throw an _ExitContext exception."""
+    raise _ExitContext
+
+
 class _Bucket:
-    """Anonymous attribute bucket class."""
+    """Anonymous attribute-bucket class."""
     pass
 
 
-class TraceManager(object):
-    """Singleton class for managing frame trace functions."""
-
-    lock = threading.Lock()
-    orig_sys_trace = None
-    orig_trace_funcs = {}
-
-    @classmethod
-    def _enable_tracing(self):
-        """Enable system-wide tracing, if it wasn't already."""
-        if len(self.orig_trace_funcs) == 1:
-            try:
-                self.orig_sys_trace = sys.gettrace()
-            except AttributeError:
-                self.orig_sys_trace = None
-            if self.orig_sys_trace is None:
-                sys.settrace(lambda *args,**kwds: None)
-
-    @classmethod
-    def _disable_tracing(self):
-        """Disable system-wide tracing, if we specifically switched it on."""
-        if len(self.orig_trace_funcs) == 1:
-            if self.orig_sys_trace is None:
-                sys.settrace(None)
-
-    @classmethod
-    def set_trace(self,frame,trace):
-        """Set a new trace function on the given frame."""
-        with self.lock:
-            self.orig_trace_funcs[frame] = frame.f_trace
-            frame.f_trace = trace
-            self._enable_tracing()
-          
-
-    @classmethod
-    def restore_trace(self,frame):
-        """Restore the original trace function to the given frame."""
-        with self.lock:
-            self._disable_tracing()
-            frame.f_trace = self.orig_trace_funcs.pop(frame)
-
-
 
 class WithHack(object):
     """Base class for with-statement-related hackery.
@@ -157,22 +120,30 @@ class WithHack(object):
         into the execution context of the with-statement.
         """
         frame = self._get_context_frame()
-        def trace(frame,event,arg):
-            frame.f_locals.update(locals)
-            TraceManager.restore_trace(frame)
-        TraceManager.set_trace(frame,trace)
+        inject_trace_func(frame,lambda frame: frame.f_locals.update(locals))
 
     def __enter__(self):
+        """Enter the context of this WithHack.
+
+        The base implementation will skip execution of the contained
+        code according to the values of "dont_execute" and "must_execute".
+        Be sure to call the superclass version if you override it.
+        """
         if self.dont_execute and not self.must_execute:
             frame = self._get_context_frame()
-            def trace(frame,event,arg):
-                raise _ExitContext
-            TraceManager.set_trace(frame,trace)
+            inject_trace_func(frame,_exit_context)
         return self
 
     def __exit__(self,exc_type,exc_value,traceback):
-        if self.dont_execute and not self.must_execute:
-            TraceManager.restore_trace(self._get_context_frame())
+        """Enter the context of this WithHack.
+
+        This is usually where all the interesting hackery takes place.
+
+        The base implementation suppresses the special _ExitContext exception
+        but lets any other exceptions pass through.  Your subclass should
+        probably do the same - the simplest way is to pass through the return
+        value given by this base implementation.
+        """
         if exc_type is _ExitContext:
             return True
         else:
@@ -205,28 +176,8 @@ class CaptureBytecode(WithHack):
 
     def __exit__(self,*args):
         frame = self._get_context_frame()
-        code = frame.f_code.co_code
-        bytecode = Code.from_code(frame.f_code)
-        offset_start = -1
-        i = 0
-        while i < self.__bc_start:
-            offset_start += 1
-            if code[i] >= HAVE_ARGUMENT:
-                i += 2
-            else:
-                i += 1
-        offset_end = offset_start
-        while i < frame.f_lasti:
-            offset_end += 1
-            if code[i] >= HAVE_ARGUMENT:
-                i += 2
-            else:
-                i += 1
-        self.bytecode_before = copy.deepcopy(bytecode)
-        self.bytecode_after = copy.deepcopy(bytecode)
-        self.bytecode_before.code = bytecode.code[:offset_start]
-        self.bytecode_after.code = bytecode.code[offset_end:]
-        bytecode.code = bytecode.code[offset_start:offset_end]
+        bytecode = extract_code(frame,self.__bc_start,frame.f_lasti)
+        #print bytecode.code
         #  Remove code setting up the with-statement block.
         while bytecode.code[0][0] != SETUP_FINALLY:
             bytecode.code = bytecode.code[1:]
@@ -235,9 +186,10 @@ class CaptureBytecode(WithHack):
         #  and remove the setup code.
         if bytecode.code[0][0] in (LOAD_FAST,LOAD_NAME,LOAD_DEREF,LOAD_GLOBAL):
             if bytecode.code[0][1].startswith("_["):
-                bytecode.code = bytecode.code[2:]
-            self.as_name = bytecode.code[0][1]
-            bytecode.code = bytecode.code[1:]
+                while bytecode.code[0][0] not in (STORE_FAST,STORE_NAME,):
+                    bytecode.code = bytecode.code[1:]
+                self.as_name = bytecode.code[0][1]
+                bytecode.code = bytecode.code[1:]
         #  Remove code tearing down the with-statement block
         while bytecode.code[-1][0] != POP_BLOCK:
             bytecode.code = bytecode.code[:-1]
@@ -330,12 +282,12 @@ class CaptureFunction(CaptureBytecode):
         return retcode
 
 
-class CaptureLocals(WithHack):
-    """WithHack to capture any local variables modified in the block.
+class CaptureLocals(CaptureBytecode):
+    """WithHack to capture any local variables assigned to in the block.
 
     When the block exits, the attribute "locals" will be a dictionary 
-    containing any local variables that were created or modified during
-    the execution of the block. 
+    containing any local variables that were assigned to during the execution
+    of the block.
 
         >>> with CaptureLocals() as f:
         ...     x = 7
@@ -349,23 +301,14 @@ class CaptureLocals(WithHack):
 
     must_execute = True
 
-    def __enter__(self):
-        frame = self._get_context_frame()
-        self.__pre_locals = frame.f_locals.copy()
-        return super(CaptureLocals,self).__enter__()
-
     def __exit__(self,*args):
+        retcode = super(CaptureLocals,self).__exit__(*args)
         frame = self._get_context_frame()
         self.locals = {}
-        for (name,value) in frame.f_locals.iteritems():
-            if value is self:
-                pass
-            elif name not in self.__pre_locals:
-                self.locals[name] = value
-            elif self.__pre_locals[name] is not value:
-                self.locals[name] = value
-        del self.__pre_locals
-        return super(CaptureLocals,self).__exit__(*args)
+        for (op,arg) in self.bytecode.code:
+           if op in (STORE_FAST,STORE_NAME,):
+               self.locals[arg] = frame.f_locals[arg]
+        return retcode
 
 
 class CaptureOrderedLocals(CaptureBytecode):
@@ -400,6 +343,47 @@ class CaptureOrderedLocals(CaptureBytecode):
         return retcode
 
 
+class CaptureModifiedLocals(WithHack):
+    """WithHack to capture any local variables modified in the block.
+
+    When the block exits, the attribute "locals" will be a dictionary 
+    containing any local variables that were created or modified during the
+    execution of the block.
+
+        >>> x = 7
+        >>> with CaptureModifiedLocals() as f:
+        ...     x = 7
+        ...     y = 8
+        ...     z = 9
+        ...
+        >>> f.locals
+        {'y': 8, 'z': 9}
+        >>>
+
+    This differs from CaptureLocals in that it does not detect variables
+    that are assigned within the block if their value doesn't actually
+    change.  It's cheaper to test for but not as reliable.
+    """
+
+    def __enter__(self):
+        frame = self._get_context_frame()
+        self.__pre_locals = frame.f_locals.copy()
+        return super(CaptureModifiedLocals,self).__enter__()
+
+    def __exit__(self,*args):
+        frame = self._get_context_frame()
+        self.locals = {}
+        for (name,value) in frame.f_locals.iteritems():
+            if value is self:
+                pass
+            elif name not in self.__pre_locals:
+                self.locals[name] = value
+            elif self.__pre_locals[name] != value:
+                self.locals[name] = value
+        del self.__pre_locals
+        return super(CaptureModifiedLocals,self).__exit__(*args)
+
+
 class xargs(CaptureOrderedLocals):
     """WithHack to call a function with arguments defined in the block.
 
@@ -515,9 +499,17 @@ class namespace(CaptureBytecode):
         funcode.code.append((LOAD_CONST,None))
         funcode.code.append((RETURN_VALUE,None))
         #  Switch LOAD/STORE/DELETE_FAST/NAME to LOAD/STORE/DELETE_ATTR
-        self._adjust_names(funcode)
+        to_replace = []
+        for (i,(op,arg)) in enumerate(funcode.code):
+            repl = self._replace_opcode((op,arg),frame)
+            if repl:
+                to_replace.append((i,repl))
+        offset = 0
+        for (i,repl) in to_replace:
+            funcode.code[i+offset:i+offset+1] = repl
+            offset += len(repl) - 1
         #  Create function object to do the manipulation
-        funcode.args = ("namespace",)
+        funcode.args = ("_[namespace]",)
         funcode.varargs = False
         funcode.varkwargs = False
         funcode.name = "<withhack>"
@@ -529,27 +521,42 @@ class namespace(CaptureBytecode):
             self._set_context_locals({self.as_name:self.namespace})
         return retcode
 
-    def _adjust_names(self,funcode):
-        for (i,(op,arg)) in enumerate(funcode.code):
-            if op in (LOAD_FAST,LOAD_NAME,):
-                funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(LOAD_ATTR,arg)]
-            elif op in (STORE_FAST,STORE_NAME,):
-                funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(STORE_ATTR,arg)]
-            elif op in (DELETE_FAST,DELETE_NAME,):
-                funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(DELETE_ATTR,arg)]
-            elif op in (LOAD_GLOBAL,LOAD_DEREF,):
-                if not self._name_used_before(arg):
-                    funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(LOAD_ATTR,arg)]
-            elif op in (DELETE_GLOBAL,):
-                if not self._name_used_before(arg):
-                    funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(DELETE_ATTR,arg)]
+    def _replace_opcode(self,(op,arg),frame):
+        if op in (STORE_FAST,STORE_NAME,):
+            return [(LOAD_FAST,"_[namespace]"),(STORE_ATTR,arg)]
+        if op in (DELETE_FAST,DELETE_NAME,):
+            return [(LOAD_FAST,"_[namespace]"),(DELETE_ATTR,arg)]
+        if op in (LOAD_FAST,LOAD_NAME,LOAD_GLOBAL,LOAD_DEREF):
+            excIn = Label(); excOut = Label(); end = Label()
+            return [(SETUP_EXCEPT,excIn),
+                        (LOAD_FAST,"_[namespace]"),(LOAD_ATTR,arg),
+                        (STORE_FAST,"_[ns_value]"),
+                        (POP_BLOCK,None),(JUMP_FORWARD,end),
+                    (excIn,None),
+                        (DUP_TOP,None),(LOAD_CONST,AttributeError),
+                        (COMPARE_OP,"exception match"),(JUMP_IF_FALSE,excOut),
+                        (POP_TOP,None),(POP_TOP,None),
+                        (POP_TOP,None),(POP_TOP,None),
+                        (LOAD_CONST,self._load_name),(LOAD_CONST,frame),
+                        (LOAD_CONST,arg),(CALL_FUNCTION,2),
+                        (STORE_FAST,"_[ns_value]"),(JUMP_FORWARD,end),
+                    (excOut,None),
+                        (POP_TOP,None),(END_FINALLY,None),
+                    (end,None),
+                        (LOAD_FAST,"_[ns_value]")]
+        return None
 
-    def _name_used_before(self,name):
-        for (op,arg) in self.bytecode_before.code:
-            if op in (LOAD_GLOBAL,LOAD_DEREF,STORE_GLOBAL,STORE_DEREF,):
-                if arg == name:
-                    return True
-        return False
+    def _load_name(self,frame,name):
+        try:
+            return frame.f_locals[name]
+        except KeyError:
+            try:
+                return frame.f_globals[name]
+            except KeyError:
+                try:
+                    return frame.f_builtins[name]
+                except KeyError:
+                    raise NameError(name)
 
 
 class keyspace(namespace):
@@ -587,18 +594,31 @@ class keyspace(namespace):
             ns = {}
         super(keyspace,self).__init__(ns)
 
-    def _adjust_names(self,funcode):
-        for (i,(op,arg)) in enumerate(funcode.code):
-            if op in (LOAD_FAST,LOAD_NAME,):
-                funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(LOAD_CONST,arg),(BINARY_SUBSCR,None)]
-            elif op in (STORE_FAST,STORE_NAME,):
-                funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(LOAD_CONST,arg),(STORE_SUBSCR,None)]
-            elif op in (DELETE_FAST,DELETE_NAME,):
-                funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(LOAD_CONST,arg),(DELETE_SUBSCR,None)]
-            elif op in (LOAD_GLOBAL,LOAD_DEREF,):
-                if not self._name_used_before(arg):
-                    funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(LOAD_CONST,arg),(BINARY_SUBSCR,None)]
-            elif op in (DELETE_GLOBAL,):
-                if not self._name_used_before(arg):
-                    funcode.code[i:i+1]=[(LOAD_FAST,"namespace"),(LOAD_CONST,arg),(DELETE_SUBSCR,None)]
+    def _replace_opcode(self,(op,arg),frame):
+        if op in (STORE_FAST,STORE_NAME,):
+            return [(LOAD_FAST,"_[namespace]"),(LOAD_CONST,arg),
+                    (STORE_SUBSCR,arg)]
+        if op in (DELETE_FAST,DELETE_NAME,):
+            return [(LOAD_FAST,"_[namespace]"),(LOAD_CONST,arg),
+                    (DELETE_SUBSCR,arg)]
+        if op in (LOAD_FAST,LOAD_NAME,LOAD_GLOBAL,LOAD_DEREF):
+            excIn = Label(); excOut = Label(); end = Label()
+            return [(SETUP_EXCEPT,excIn),
+                        (LOAD_FAST,"_[namespace]"),(LOAD_CONST,arg),
+                        (BINARY_SUBSCR,arg),
+                        (STORE_FAST,"_[ns_value]"),
+                        (POP_BLOCK,None),(JUMP_FORWARD,end),
+                    (excIn,None),
+                        (DUP_TOP,None),(LOAD_CONST,KeyError),
+                        (COMPARE_OP,"exception match"),(JUMP_IF_FALSE,excOut),
+                        (POP_TOP,None),(POP_TOP,None),
+                        (POP_TOP,None),(POP_TOP,None),
+                        (LOAD_CONST,self._load_name),(LOAD_CONST,frame),
+                        (LOAD_CONST,arg),(CALL_FUNCTION,2),
+                        (STORE_FAST,"_[ns_value]"),(JUMP_FORWARD,end),
+                    (excOut,None),
+                        (POP_TOP,None),(END_FINALLY,None),
+                    (end,None),
+                        (LOAD_FAST,"_[ns_value]")]
+        return None
 
